package dk.kapetanovic.devicepolicymanager;

import android.app.admin.DevicePolicyManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;

import org.apache.cordova.CallbackContext;
import org.apache.cordova.CordovaInterface;
import org.apache.cordova.CordovaPlugin;
import org.apache.cordova.CordovaWebView;
import org.apache.cordova.PluginResult;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.List;

public class DevicePolicyManagerPlugin extends CordovaPlugin {
    private static final int REQUEST_CODE_ADD_DEVICE_ADMIN = 0;

    private static final String ACTION_ADD_PERSISTED_PREFERRED_ACTIVITY = "addPersistentPreferredActivity";
    private static final String ACTION_ADD_USER_RESTRICTION = "addUserRestriction";
    private static final String ACTION_CLEAR_DEVICE_OWNER_APP = "clearDeviceOwnerApp";
    private static final String ACTION_CLEAR_PACKAGE_PERSISTENT_PREFERRED_ACTIVITIES = "clearPackagePersistentPreferredActivities";
    private static final String ACTION_GET_ACTIVE_ADMINS = "getActiveAdmins";
    private static final String ACTION_GET_PERMISSION_GRANT_STATE = "getPermissionGrantState";
    private static final String ACTION_GET_PERMISSION_POLICY = "getPermissionPolicy";
    private static final String ACTION_IS_ADMIN_ACTIVE = "isAdminActive";
    private static final String ACTION_IS_DEVICE_OWNER_APP = "isDeviceOwnerApp";
    private static final String ACTION_IS_LOCK_TASK_PERMITTED = "isLockTaskPermitted";
    private static final String ACTION_IS_PROFILE_OWNER_APP = "isProfileOwnerApp";
    private static final String ACTION_LOCK_NOW = "lockNow";
    private static final String ACTION_REMOVE_ACTIVE_ADMIN = "removeActiveAdmin";
    private static final String ACTION_SET_GLOBAL_SETTING = "setGlobalSetting";
    private static final String ACTION_SET_LOCK_TASK_PACKAGES = "setLockTaskPackages";
    private static final String ACTION_SET_KEYGUARD_DISABLED = "setKeyguardDisabled";
    private static final String ACTION_SET_PERMISSION_GRANT_STATE = "setPermissionGrantState";
    private static final String ACTION_SET_PERMISSION_POLICY = "setPermissionPolicy";
    private static final String ACTION_SET_STATUS_BAR_DISABLED = "setStatusBarDisabled";
    private static final String ACTION_ON_CHANGE = "onChange";

    private ComponentName adminComponentName;
    private DevicePolicyManager devicePolicyManager;
    private final List<CallbackClosure> addDeviceAdminCallbacks = new ArrayList<CallbackClosure>();

    @Override
    public void initialize (CordovaInterface cordova, CordovaWebView webView) {
        super.initialize(cordova, webView);

        adminComponentName = new ComponentName(webView.getContext(), DeviceAdminReceiverImpl.class);
        devicePolicyManager = (DevicePolicyManager) cordova.getActivity()
                .getApplication()
                .getApplicationContext()
                .getSystemService(Context.DEVICE_POLICY_SERVICE);
    }

    @Override
    public boolean execute (String action, JSONArray args, CallbackContext callbackContext) throws JSONException {
        if (action.equals(ACTION_ADD_PERSISTED_PREFERRED_ACTIVITY)) addPersistentPreferredActivity(args, callbackContext);
        else if (action.equals(ACTION_ADD_USER_RESTRICTION)) addUserRestriction(args, callbackContext);
        else if (action.equals(ACTION_CLEAR_DEVICE_OWNER_APP)) clearDeviceOwnerApp(callbackContext);
        else if (action.equals(ACTION_CLEAR_PACKAGE_PERSISTENT_PREFERRED_ACTIVITIES)) clearPackagePersistentPreferredActivities(args, callbackContext);
        else if (action.equals(ACTION_GET_ACTIVE_ADMINS)) getActiveAdmins(callbackContext);
        else if (action.equals(ACTION_GET_PERMISSION_GRANT_STATE)) getPermissionGrantState(args, callbackContext);
        else if (action.equals(ACTION_GET_PERMISSION_POLICY)) getPermissionPolicy(callbackContext);
        else if (action.equals(ACTION_IS_ADMIN_ACTIVE)) isAdminActive(callbackContext);
        else if (action.equals(ACTION_IS_DEVICE_OWNER_APP)) isDeviceOwnerApp(args, callbackContext);
        else if (action.equals(ACTION_IS_LOCK_TASK_PERMITTED)) isLockTaskPermitted(args, callbackContext);
        else if (action.equals(ACTION_IS_PROFILE_OWNER_APP)) isProfileOwnerApp(args, callbackContext);
        else if (action.equals(ACTION_LOCK_NOW)) lockNow(callbackContext);
        else if (action.equals(ACTION_REMOVE_ACTIVE_ADMIN)) removeActiveAdmin(callbackContext);
        else if (action.equals(ACTION_SET_GLOBAL_SETTING)) setGlobalSetting(args, callbackContext);
        else if (action.equals(ACTION_SET_LOCK_TASK_PACKAGES)) setLockTaskPackages(args, callbackContext);
        else if (action.equals(ACTION_SET_KEYGUARD_DISABLED)) setKeyguardDisabled(args, callbackContext);
        else if (action.equals(ACTION_SET_PERMISSION_GRANT_STATE)) setPermissionGrantState(args, callbackContext);
        else if (action.equals(ACTION_SET_PERMISSION_POLICY)) setPermissionPolicy(args, callbackContext);
        else if (action.equals(ACTION_SET_STATUS_BAR_DISABLED)) setStatusBarDisabled(args, callbackContext);
        else if (action.equals(ACTION_ON_CHANGE)) onChange(callbackContext);
        else return false;

        return true;
    }

    @Override
    public void onActivityResult (int requestCode, int resultCode, Intent intent) {
        if (requestCode == REQUEST_CODE_ADD_DEVICE_ADMIN) {
            boolean isActive = isAdminActive();

            synchronized (addDeviceAdminCallbacks) {
                for (CallbackClosure callbackClosure : addDeviceAdminCallbacks) {
                    try {
                        if (isActive) {
                            callbackClosure.execute();
                        } else {
                            CallbackContext callbackContext = callbackClosure.getCallbackContext();
                            callbackContext.sendPluginResult(ERROR("Device admin not active"));
                        }
                    } catch (JSONException e) {
                        CallbackContext callbackContext = callbackClosure.getCallbackContext();
                        PluginResult result = new PluginResult(PluginResult.Status.JSON_EXCEPTION, e.getMessage());
                        callbackContext.sendPluginResult(result);
                    }
                }

                addDeviceAdminCallbacks.clear();
            }
        }
    }

    private void addPersistentPreferredActivity (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                JSONObject json = args.getJSONObject(0);
                IntentFilter intentFilter = fromJSONIntentFilter(json);
                String activity = args.getString(1);
                ComponentName componentName = ComponentName.unflattenFromString(activity);

                devicePolicyManager.addPersistentPreferredActivity(adminComponentName, intentFilter, componentName);
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void addUserRestriction (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                String key = args.getString(0);
                devicePolicyManager.addUserRestriction(adminComponentName, key);
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void clearDeviceOwnerApp (CallbackContext callbackContext) throws JSONException {
        String pkg = cordova.getActivity().getPackageName();
        devicePolicyManager.clearDeviceOwnerApp(pkg);
        callbackContext.sendPluginResult(OK(null));
    }

    private void clearPackagePersistentPreferredActivities (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                String pkg = args.getString(0);
                devicePolicyManager.clearPackagePersistentPreferredActivities(adminComponentName, pkg);
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void getActiveAdmins (CallbackContext callbackContext) throws JSONException {
        List<ComponentName> admins = devicePolicyManager.getActiveAdmins();
        JSONArray json = new JSONArray();

        if (admins != null) {
            for (ComponentName componentName : admins) {
                json.put(componentName.flattenToString());
            }
        }

        callbackContext.sendPluginResult(OK(json));
    }

    private void getPermissionGrantState (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                String pkg = args.getString(0);
                String permission = args.getString(1);
                int result = devicePolicyManager.getPermissionGrantState(adminComponentName, pkg, permission);
                callbackContext.sendPluginResult(OK(toStringPermissionGrantState(result)));
            }
        });
    }

    private void getPermissionPolicy (final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                int result = devicePolicyManager.getPermissionPolicy(adminComponentName);
                callbackContext.sendPluginResult(OK(toStringPermissionPolicy(result)));
            }
        });
    }

    private void isAdminActive (CallbackContext callbackContext) throws JSONException {
        boolean active = isAdminActive();
        callbackContext.sendPluginResult(OK(active));
    }

    private void isDeviceOwnerApp (JSONArray args, CallbackContext callbackContext) throws JSONException {
        String pkg = args.getString(0);
        boolean result = devicePolicyManager.isDeviceOwnerApp(pkg);
        callbackContext.sendPluginResult(OK(result));
    }

    private void isLockTaskPermitted (JSONArray args, CallbackContext callbackContext) throws JSONException {
        String pkg = args.getString(0);
        boolean result = devicePolicyManager.isLockTaskPermitted(pkg);
        callbackContext.sendPluginResult(OK(result));
    }

    private void isProfileOwnerApp (JSONArray args, CallbackContext callbackContext) throws JSONException {
        String pkg = args.getString(0);
        boolean result = devicePolicyManager.isProfileOwnerApp(pkg);
        callbackContext.sendPluginResult(OK(result));
    }

    private void lockNow (final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                devicePolicyManager.lockNow();
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void removeActiveAdmin (CallbackContext callbackContext) throws JSONException {
        devicePolicyManager.removeActiveAdmin(adminComponentName);
        callbackContext.sendPluginResult(OK(null));
    }

    private void setGlobalSetting (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                String setting = args.getString(0);
                String value = args.getString(1);
                devicePolicyManager.setGlobalSetting(adminComponentName, setting, value);
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void setLockTaskPackages (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                JSONArray jsonPackages = args.getJSONArray(0);
                String[] arrayPackages = new String[jsonPackages.length()];

                for (int i = 0; i < jsonPackages.length(); i++) {
                    arrayPackages[i] = jsonPackages.getString(i);
                }

                devicePolicyManager.setLockTaskPackages(adminComponentName, arrayPackages);
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void setKeyguardDisabled (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                boolean disabled = args.getBoolean(0);
                boolean result = devicePolicyManager.setKeyguardDisabled(adminComponentName, disabled);
                callbackContext.sendPluginResult(OK(result));
            }
        });
    }

    private void setPermissionGrantState (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                String pkg = args.getString(0);
                String permission = args.getString(1);
                int grantState = fromStringPermissionGrantState(args.getString(2));

                boolean result = devicePolicyManager.setPermissionGrantState(adminComponentName, pkg, permission,grantState);
                callbackContext.sendPluginResult(OK(result));
            }
        });
    }

    private void setPermissionPolicy (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                int policy = fromStringPermissionPolicy(args.getString(0));
                devicePolicyManager.setPermissionPolicy(adminComponentName, policy);
                callbackContext.sendPluginResult(OK(null));
            }
        });
    }

    private void setStatusBarDisabled (final JSONArray args, final CallbackContext callbackContext) throws JSONException {
        ensureAdminActive(new CallbackClosure(callbackContext) {
            @Override
            public void execute () throws JSONException {
                boolean disabled = args.getBoolean(0);
                boolean result = devicePolicyManager.setStatusBarDisabled(adminComponentName, disabled);
                callbackContext.sendPluginResult(OK(result));
            }
        });
    }

    private void onChange(CallbackContext callbackContext) {
        DeviceAdminReceiverImpl.onChange = callbackContext;
    }

    private void ensureAdminActive(CallbackClosure callbackClosure) throws JSONException {
        if (isAdminActive()) {
            callbackClosure.execute();
        } else {
            synchronized (addDeviceAdminCallbacks) {
                if (isAdminActive()) {
                    callbackClosure.execute();
                    return;
                }

                addDeviceAdminCallbacks.add(callbackClosure);

                if (addDeviceAdminCallbacks.size() == 1) {
                    Intent intent = new Intent(DevicePolicyManager.ACTION_ADD_DEVICE_ADMIN);
                    intent.putExtra(DevicePolicyManager.EXTRA_DEVICE_ADMIN, adminComponentName);
                    intent.putExtra(DevicePolicyManager.EXTRA_ADD_EXPLANATION, "Enable device admin for this application");
                    cordova.startActivityForResult(this, intent, REQUEST_CODE_ADD_DEVICE_ADMIN);
                }
            }
        }
    }

    private boolean isAdminActive () {
        return devicePolicyManager.isAdminActive(adminComponentName);
    }

    private static IntentFilter fromJSONIntentFilter (JSONObject json) throws JSONException {
        IntentFilter intentFilter = new IntentFilter();

        JSONArray actions = json.optJSONArray("actions");
        JSONArray categories = json.optJSONArray("categories");

        if (actions != null) {
            for (int i = 0; i < actions.length(); i++) {
                intentFilter.addAction(actions.getString(i));
            }
        }

        if (categories != null) {
            for (int i = 0; i < categories.length(); i++) {
                intentFilter.addCategory(categories.getString(i));
            }
        }

        return intentFilter;
    }

    private static String toStringPermissionGrantState (int state) {
        switch (state) {
            case DevicePolicyManager.PERMISSION_GRANT_STATE_DEFAULT: return "DEFAULT";
            case DevicePolicyManager.PERMISSION_GRANT_STATE_GRANTED: return "GRANTED";
            case DevicePolicyManager.PERMISSION_GRANT_STATE_DENIED: return "DENIED";
            default: return null;
        }
    }

    private static String toStringPermissionPolicy (int policy) {
        switch (policy) {
            case DevicePolicyManager.PERMISSION_POLICY_PROMPT: return "PROMPT";
            case DevicePolicyManager.PERMISSION_POLICY_AUTO_GRANT: return "AUTO_GRANT";
            case DevicePolicyManager.PERMISSION_POLICY_AUTO_DENY: return "AUTO_DENY";
            default: return null;
        }
    }

    private static int fromStringPermissionGrantState (String state) {
        if (state.equals("DEFAULT")) return DevicePolicyManager.PERMISSION_GRANT_STATE_DEFAULT;
        if (state.equals("GRANTED")) return DevicePolicyManager.PERMISSION_GRANT_STATE_GRANTED;
        if (state.equals("DENIED")) return DevicePolicyManager.PERMISSION_GRANT_STATE_DENIED;
        return -1;
    }

    private static int fromStringPermissionPolicy (String policy) {
        if (policy.equals("PROMPT")) return DevicePolicyManager.PERMISSION_POLICY_PROMPT;
        if (policy.equals("AUTO_GRANT")) return DevicePolicyManager.PERMISSION_POLICY_AUTO_GRANT;
        if (policy.equals("AUTO_DENY")) return DevicePolicyManager.PERMISSION_POLICY_AUTO_DENY;
        return -1;
    }

    private static PluginResult OK (Object obj) throws JSONException {
        return createPluginResult(obj, PluginResult.Status.OK);
    }

    private static PluginResult ERROR (Object obj) throws JSONException {
        return createPluginResult(obj, PluginResult.Status.ERROR);
    }

    private static PluginResult createPluginResult (Object obj, PluginResult.Status status) throws JSONException {
        JSONObject json = new JSONObject();
        json.put("data", obj == null ? JSONObject.NULL : obj);
        return new PluginResult(status, json);
    }

    private abstract class CallbackClosure {
        private CallbackContext callbackContext;

        public CallbackClosure (CallbackContext callbackContext) {
            this.callbackContext = callbackContext;
        }

        public CallbackContext getCallbackContext () {
            return callbackContext;
        }

        abstract public void execute () throws JSONException;
    }
}
